/*
  Uložená procedura vrátí unikátní číslo z dané řady. Řada je typicky určena pro konkrétní modul a je alespoň jedna.
  Vstupní parametry:
  * i_idDokl         - číslo dokladu (id) 
  * i_idRady         - řada, která se má použít pro vygenerování čísla
  * i_idUcetniObdobi - id účetního období
  * i_Modul          - zkratka modulu, do kteréhoho doklad spadá
  * i_TypPohybuK     - typ pohybu (příjem/výdej). V současné době se nepoužívá.

  Jak přidělování čísla funguje
  ==============================
  - pokud příslušná řada pro daný rok neexistuje, je vytvořena. Pokud existuje řada pro nějaký předchozí rok, vezme se
    nastavení z ní. V opačném případě se vytvoří dle výchozích hodnot tak, aby byla v rámci dvojice kód, modul unikátní. 
    Výchozí hodnoty pro řadu: i_idRady,i_idUcetObdobi,8,1,1,1,'','/%YY%',now()  
  - při získávání čísla se nejprve podíváme do seznamu děr (dRadyDiry), zda neexistuje nějaké číslo, které si klient
    vyžádal, ale navrátil ho. Vyhledáme si zde všechny čísla starší dvou hodin, po kterou má klient číslo "rezervováno"
    (viz dále). Pokud je zde volné číslo, vrátíme ho a zvýšíme čas rezervace o dvě hodiny (LastUpdate:=AktČas+2 hodiny).    
  - pokud není žádné číslo v seznamu děr, vygenerujeme nové číslo a v tabulce řady zvýšíme AktCis o jedničku. Ještě
    před navrácením čísla klientovi ho uložíme do seznamu děr jako "rezervaci", což znamená, že ho uložíme s 
    LastUpdate:=AktČas+2 hodiny. !!! AktCis je číslo udávající naposledy přidělené číslo !!!
  - číslo je před navrácením zformátováno podle formátu čís. řady pro daný kalendářní rok.
  
  Možné scénáře, které mohou nastat
  ==================================
  1) Klient zažádá o číslo, to je jako "rezervace" vloženo do tabulky děr a dáno klientu. Ten doklad, pro který bylo
     číslo vygenerováno, vymaže ze seznamu děr. Vše OK.
  2) Klient zažádá o číslo, to je jako "rezervace" vloženo do tabulky děr a dáno klientu. Ten vytváření dokladu
     stornuje, aplikace korektně navrátí číslo. Pokud je toto číslo na vrcholu přidělených čísel tj. nebylo v číselné
     řadě přiděleno další číslo, tak se provede snížení čítače (AktCis) o jedničku a vymaže se "rezervace" z tabulky děr.
     Pokud již bylo přiděleno novější číslo, je "rezervace" v tabulce děr označena aktuálním datem, což znamená, že jej
     může získat libovolný klient, který zažádá o přidělení čísla.
  3) Klient zažádá o číslo, to je jako "rezervace" vloženo do tabulky děr a dáno klientu. Ten z nějakého důvodu přestane
     reagovat a nenavrátí přidělené číslo. V tom případě vygenerované číslo v seznamu děr "vytiká" a po dvou hodinách
     je opět dostupné libovolnému klientu.
  4) Klient zažádá o číslo, to je jako "rezervace" vloženo do tabulky děr a dáno klientu. Klient odejde na oběd a vrátí
     se po více jak dvou hodinách. Mezitím bylo po dvou hodinách uvedené číslo přiděleno někomu jinému. V takovém případě
     se stejně jako v předchozích případech při insertu v triggeru zkontroluje, zda uvedené číslo již náhodou někdo nemá
     v tabulce s příslušnými doklady. Pokud má, vygeneruje se klientu nové číslo.
     
  Na co je potřeba brát zřetel
  =============================
  - řada obsahuje zástupné znaky "%Y%" až "%YYYY", které jsou nahrazovány příšlusnou částí roku předaného účetního období.
    Tím je zajištěno, že každá řada pro každý rok je unikátní. Nerozlišuje však řady se stejným formátem pr ostejný modul!
  - pokud uvedené zástupné znaky nejsou, nastává problém, kdy nelze od nového roku začít od nuly, protože by se to tlouklo
    se stejnými čísly dokladů z loňského roku (např. "A00001" za rok 2007 vs "A00002" za rok 2008). Pokud tato situace
    nastane, vezme se nejvyšší dosud přidělené číslo v předchozím roce a přidělí se číslo další. Toto přidělování je shora
    omezeno číslem 100 000.
  
  Verze pro PostgreSQL:     Pavel Cvrček
  Původní verze pro Sybase: Josef Weinreb
*/

create or replace function dGetCisDokl 
(in i_idDokl integer, in i_idRady integer, in i_idUcetniObdobi integer, in i_Modul varchar(3), in i_TypPohybuK varchar(50))
returns refcursor as $$
declare
  _idRadyRok integer;
  _CisDelka integer;
  _ZobrazNuly boolean;
  _CisAkt integer;
  _CisPoc integer;
  _Prefix varchar(8);
  _Postfix varchar(8);
  _PocetZaznamu integer;
  
  _idRadyDiry integer;
  _CisDira varchar(20);  
  _Unique boolean;
  _iter integer;

  o_Chyba integer;
  o_CisDokl varchar(20);
  
  _navratoveHodnoty refcursor;
begin
  o_Chyba:= 0;
  o_CisDokl:= null;
  
  /* Ověření, zda pro daný účetní rok existuje dokladová řada pro rok. Pokud neexistuje, vytvoří se. */
  select idRadyRok into _idRadyRok from dRadyRok where idRady = i_idRady and iducetobdobi = i_idUcetniObdobi;
  if(_idRadyRok is null) then
    execute dZalozRaduRok (i_idRady, i_idUcetniObdobi, i_Modul);
  end if;
  
  /* Načte se dokladová řada pro účetní rok */
  select CisDelka, ZobrazNuly, CisAkt, CisPoc, Prefix, Postfix, idRadyRok into _CisDelka,_ZobrazNuly,_CisAkt,_CisPoc,_Prefix,_Postfix,
    _idRadyRok from dRadyRok where idRady = i_idRady and iducetobdobi = i_idUcetniObdobi for update;      


  /* Podíváme se do seznamu děr, zda tam náhodou neexistuje vrácené číslo. Pokud ano, zjistíme jeho unikátnost. Pokud je unikátní,
     provedeme jeho přidělení klientu. V opačném případě jej smažeme a pokusíme se vybrat nové číslo ze seznamu děr. */
  select CisDira, idRadyDiry into _CisDira,_idRadyDiry from dRadyDiry where idRadyRok = _idRadyRok and LastUpdate <= now() order by cisdira asc limit 1;
  if (_CisDira is not null) then
    
    _Unique:= (select dUniqueCisDokl(-1,i_Modul,i_TypPohybuK,_CisDira));    
    while(_Unique = false and _CisDira is not null) loop
      delete from dRadyDiry where idRadyDiry = _idRadyDiry;
      
      select CisDira, idRadyDiry into _CisDira,_idRadyDiry from dRadyDiry where idRadyRok = _idRadyRok and LastUpdate <= now() order by CisDira asc limit 1;
      if(_CisDira is not null) then
        _Unique:= (select dUniqueCisDokl(-1,i_Modul,i_TypPohybuK,_CisDira));
      end if;
      
    end loop;

    /* Pokud bylo nalezeno unikátní číslo, provede se jeho rezervace a vrátí se klientu. */
    if(_Unique = true and _CisDira is not null) then
      o_CisDokl:= _CisDira;
      update dRadyDiry set LastUpdate = now() + interval '2 hours' where idRadyDiry = _idRadyDiry;
    end if;      
  end if;
  
  /* Pokud nebylo číslo vzato z děr, provede se jeho přidělení standardním způsobem z řady.
     Pozor! Pokud se jedná o první číslo řady, je potřeba vrátit aktuální číslo a nikoliv inkrementované. */
  if (o_CisDokl is null) then    
    if (_CisAkt <> _CisPoc) then
      _CisAkt:= _CisAkt+1;
    end if;
    
    /* Získá se číslo a provede se test na jeho unikátnost.*/
    o_CisDokl:= (select dVytvorCisDokl(_Prefix,_CisAkt,_Postfix,_ZobrazNuly,_CisDelka,i_idUcetniObdobi));
    _Unique:= (select dUniqueCisDokl(-1,i_Modul,i_TypPohybuK,o_CisDokl));
    
    /* Podíváme se také do děr, protože počátek řady mohl být již vyžádán a je umístěn v dírách.
       Typickým příkladem je generování závazků u Leasingu, kde se vytahuje několik čísel najednou. */
    if (_Unique = true) then
      select count(*) into _PocetZaznamu from dRadyDiry where cisdira=o_CisDokl and idradyrok=_idRadyRok;
      if (_PocetZaznamu > 0) then
        _Unique:= false;
      end if;
    end if;
    
    /* Pokud není číslo unikátní, pokus se vygenerovat nové. */
    _iter:= 0;
    while (_Unique = false) loop
      _CisAkt:= _CisAkt+1;
      _iter:= _iter + 1;
      
      o_CisDokl:= (select dVytvorCisDokl(_Prefix,_CisAkt,_Postfix,_ZobrazNuly,_CisDelka,i_idUcetniObdobi));
      _Unique:=(select dUniqueCisDokl(-1,i_Modul,i_TypPohybuK,o_CisDokl));    
      
      if (_iter > 100000) then
        /* Číslo se z neznámých důvodů nepodařilo vygenerovat. Pozn.: Ochrana proti zacyklení.*/
        raise exception 'dGetCisDokl(): Číslo dokladu se nepodařilo vygenerovat.';
      end if;
    end loop;
    
    /* Nastavení chybových stavů (pokud nějaké nastaly). */
    if (o_CisDokl = 'x') then /* Kapacita číselné řady vyčerpána. */
      o_Chyba:= 1;
    end if;
    if (o_CisDokl = 'xx') then /* Délka číselné řady je delší. Max. 20 znaků.  */
      o_Chyba:= 2; 
    end if;
    if (o_CisDokl = 'xxx') then /* Maximální délka čísla smí být 12 znaků. */
      o_Chyba:= 3;
    end if;     
    
    /* Podařilo se vygenerovat unikátní číslo, provede se jeho rezervace a nastaví se nový vrchol dokladové řady. */
    if (o_Chyba = 0) and (select count(*) from dRadyDiry where idRadyRok = _idRadyRok and CisDira = o_CisDokl) = 0 then
      insert into dRadyDiry(idRadyRok,CisDira,LastUpdate) values(_idRadyRok,o_CisDokl, now() + interval '2 hours');
      update dRadyRok set cisAkt = _cisAkt where idRadyRok = _idRadyRok;
    end if;
        
  end if;   
  
  open _navratoveHodnoty for select o_CisDokl as CisloDokladu, o_Chyba as Cislo;
  return _navratoveHodnoty;  
end;
$$ LANGUAGE plpgsql;
ALTER FUNCTION public.dGetCisDokl(in i_idDokl integer, in i_idRady integer, in i_idUcetniObdobi integer, in i_Modul varchar(3), in i_TypPohybuK varchar(50)) OWNER TO %firmaRole%;
